import pandas as pd
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class SplitStep(BaseStep):
    name = Field('split', const=True)
    column: ColumnName
    delimiter: str
    # at least one col to keep
    number_cols_to_keep: int = Field(gt=0)

    def execute(
        self,
        df: pd.DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> pd.DataFrame:
        all_new_col = {}
        new_serie = df[self.column].str.split(self.delimiter)
        for i in range(self.number_cols_to_keep):
            all_new_col[f'{self.column}_{i + 1}'] = new_serie.str[i]
        return df.assign(**all_new_col)



from pandas import DataFrame
from pydantic import Field

from weaverbird.formula import clean_formula, eval_formula
from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName


class FormulaStep(BaseStep):
    name = Field('formula', const=True)
    new_column: ColumnName
    formula: str

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        serie = eval_formula(df, clean_formula(self.formula))
        return df.assign(**{self.new_column: serie})



from typing import List, Union

from pydantic import BaseModel

from weaverbird.steps import (
    AddMissingDatesStep,
    AggregateStep,
    AppendStep,
    ArgmaxStep,
    ArgminStep,
    ConcatenateStep,
    ConvertStep,
    CumSumStep,
    DateExtractStep,
    DeleteStep,
    DomainStep,
    DuplicateStep,
    DurationStep,
    EvolutionStep,
    FillnaStep,
    FilterStep,
    FormulaStep,
    FromdateStep,
    IfthenelseStep,
    JoinStep,
    LowercaseStep,
    MovingAverageStep,
    PercentageStep,
    PivotStep,
    RankStep,
    RenameStep,
    ReplaceStep,
    RollupStep,
    SelectStep,
    SortStep,
    SplitStep,
    StatisticsStep,
    SubstringStep,
    TextStep,
    ToDateStep,
    TopStep,
    TotalsStep,
    UniqueGroupsStep,
    UnpivotStep,
    UppercaseStep,
    WaterfallStep,
)

PipelineStep = Union[
    AddMissingDatesStep,
    AggregateStep,
    AppendStep,
    ArgmaxStep,
    ArgminStep,
    ConcatenateStep,
    ConvertStep,
    CumSumStep,
    DateExtractStep,
    DeleteStep,
    DomainStep,
    DuplicateStep,
    DurationStep,
    EvolutionStep,
    FillnaStep,
    FilterStep,
    FromdateStep,
    FormulaStep,
    JoinStep,
    RankStep,
    RenameStep,
    PercentageStep,
    StatisticsStep,
    IfthenelseStep,
    FromdateStep,
    LowercaseStep,
    SelectStep,
    PivotStep,
    ReplaceStep,
    SortStep,
    TextStep,
    UnpivotStep,
    TopStep,
    ToDateStep,
    RollupStep,
    UniqueGroupsStep,
    UppercaseStep,
    SplitStep,
    SubstringStep,
    WaterfallStep,
    TotalsStep,
    MovingAverageStep,
]


class Pipeline(BaseModel):
    steps: List[PipelineStep]



from typing import List, Literal

from pandas import DataFrame
from pydantic import BaseModel, Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor, PopulatedWithFieldnames


class ColumnSort(BaseModel):
    class Config(PopulatedWithFieldnames):
        ...

    column: ColumnName
    order: Literal['asc', 'desc']


class SortStep(BaseStep):
    name = Field('sort', const=True)
    columns: List[ColumnSort]

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        return df.sort_values(
            by=[sort.column for sort in self.columns],
            ascending=[sort.order == 'asc' for sort in self.columns],
        )



from typing import Any, Union

import numpy
from pandas import DataFrame
from pydantic import BaseModel, Field

from weaverbird.conditions import Condition
from weaverbird.formula import clean_formula, eval_formula
from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor, PopulatedWithFieldnames


class IfThenElse(BaseModel):
    class Config(PopulatedWithFieldnames):
        ...

    condition: Condition = Field(alias='if')
    then: str
    else_value: Union['IfThenElse', Any] = Field(alias='else')

    def execute_ifthenelse(self, df, new_column):
        if isinstance(self.else_value, IfThenElse):
            else_branch = self.else_value.execute_ifthenelse(df, new_column)[new_column]
        else:
            else_branch = eval_formula(df, clean_if_formula(self.else_value))

        then_branch = eval_formula(df, clean_if_formula(self.then))

        return df.assign(
            **{new_column: numpy.where(self.condition.filter(df), then_branch, else_branch)}
        )


IfThenElse.update_forward_refs()


class IfthenelseStep(BaseStep, IfThenElse):
    class Config(PopulatedWithFieldnames):
        ...

    name = Field('ifthenelse', const=True)
    new_column: ColumnName = Field(alias='newColumn')

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        return self.execute_ifthenelse(df, self.new_column)


# ifthenelse can take as a parameter either a formula, or a value
def clean_if_formula(formula_or_value: Any) -> Any:
    if isinstance(formula_or_value, str):
        return clean_formula(formula_or_value)
    else:
        return formula_or_value



from typing import List

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class SelectStep(BaseStep):
    name = Field('select', const=True)
    columns: List[ColumnName] = Field(min_items=1)

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        return df[self.columns]



from typing import List, Optional

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class MovingAverageStep(BaseStep):
    name = Field('movingaverage', const=True)

    value_column: Optional[ColumnName] = Field(None, alias='valueColumn')
    column_to_sort: ColumnName = Field(alias='columnToSort')
    moving_window: int = Field(alias='movingWindow')
    groups: List[ColumnName] = []
    new_column_name: Optional[ColumnName] = Field(alias='newColumnName')

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        new_column_name = self.new_column_name or f'{self.value_column}_MOVING_AVG'
        df = df.sort_values(by=self.groups + [self.column_to_sort]).reset_index(drop=True)
        if self.groups:
            df_grouped = df.groupby(self.groups)
        else:
            df_grouped = df
        serie = (
            df_grouped.rolling(self.moving_window).mean()[self.value_column].reset_index(drop=True)
        )
        return df.assign(**{new_column_name: serie})



from typing import List

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class PercentageStep(BaseStep):
    name = Field('percentage', const=True)
    column: ColumnName
    group: List[ColumnName] = Field(default=[])
    new_column_name: ColumnName = Field(alias='newColumnName', default=None)

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        new_column_name = self.new_column_name or f'{self.column}_PCT'

        if len(self.group) > 0:
            sums = df.groupby(self.group)[self.column].transform('sum')
        else:
            sums = df[self.column].sum()
        return df.assign(**{new_column_name: df[self.column] / sums})



from typing import List, Optional

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName


class CumSumStep(BaseStep):
    name = Field('cumsum', const=True)
    value_column: ColumnName = Field(..., alias='valueColumn')
    reference_column: ColumnName = Field(..., alias='referenceColumn')
    groupby: Optional[List[ColumnName]]
    new_column: Optional[ColumnName] = Field(None, alias='newColumn')

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        df = df.sort_values(self.reference_column)
        dst_column = self.new_column or f'{self.value_column}_CUMSUM'
        df_grouped = df.groupby(self.groupby) if self.groupby else df
        cumsum_serie = df_grouped[self.value_column].cumsum()
        return df.assign(**{dst_column: cumsum_serie})



import re

from pandas import DataFrame

COLUMN_PATTERN = re.compile(r'(\`[^\`]*\`)')


# In pandas, columns containing spaces must be escaped with backticks:
def clean_formula_element(elem: str) -> str:
    if COLUMN_PATTERN.match(elem):
        # this is a column name: return it as is
        return elem
    # - forbid '=' char
    # - replace [ and ] with `
    #   ([] is mongo's syntax to escape columns containing spaces)
    return elem.replace('=', '').replace('[', '`').replace(']', '`')


def clean_formula(formula: str) -> str:
    """
    Translate mongo's syntax to hande columns names containing spaces to pandas syntax.

    Example:
        >>> clean_formula('colA * `col B` * [col C] * `[col D]`')
        'colA * `col B` * `col C` * `[col D]`'
    """
    formula_splitted = COLUMN_PATTERN.split(formula)
    return ''.join(clean_formula_element(elem) for elem in formula_splitted)


def eval_formula(df: DataFrame, formula: str) -> DataFrame:
    try:
        return df.eval(formula)
    except Exception:
        # for all cases not handled by NumExpr
        return df.eval(formula, engine='python')



from typing import Any, List, Literal, Optional

import numpy
import pandas as pd
from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor

TYPE_WATERFALL_COLUMN = 'TYPE_waterfall'
LABEL_WATERFALL_COLUMN = 'LABEL_waterfall'
GROUP_WATERFALL_COLUMN = 'GROUP_waterfall'

_RESULT_COLUMN = 'result'


class WaterfallStep(BaseStep):
    name = Field('waterfall', const=True)
    valueColumn: ColumnName
    milestonesColumn: ColumnName
    start: Any
    end: Any
    labelsColumn: ColumnName
    sortBy: Literal['value', 'label']
    order: Literal['desc', 'asc']
    parentsColumn: Optional[ColumnName]
    groupby: List[ColumnName] = []

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        # first milestone
        start_df = df[df[self.milestonesColumn] == self.start].rename(
            columns={self.valueColumn: f'{self.valueColumn}_start'}
        )
        # second milestone
        end_df = df[df[self.milestonesColumn] == self.end].rename(
            columns={self.valueColumn: f'{self.valueColumn}_end'}
        )

        upper = self.compute_agg_milestone(
            df, start_df.rename(columns={f'{self.valueColumn}_start': self.valueColumn}), self.start
        )
        mid = self.compute_mid(self.merge(start_df, end_df), df)
        downer = self.compute_agg_milestone(
            df, end_df.rename(columns={f'{self.valueColumn}_end': self.valueColumn}), self.end
        )
        return pd.concat([upper, mid, downer])

    # start_df is the base dataframe filtered to contains only result at the start of the waterfall
    # end_df is the base dataframe filtered to contains only result at the start of the waterfall
    # this methods compute the difference for the value value of every label between the end and the start
    def merge(self, start_df, end_df):

        group_by_columns = [self.labelsColumn] + self.groupby
        if self.parentsColumn is not None:
            group_by_columns = group_by_columns + [self.parentsColumn]
        start_df = (
            start_df.groupby(by=group_by_columns)
            .agg({self.valueColumn + '_start': 'sum'})
            .merge(start_df[group_by_columns], on=group_by_columns)
            .rename(columns={self.labelsColumn + '_x': self.labelsColumn})
            .drop_duplicates()
        )

        end_df = (
            end_df.groupby(by=group_by_columns)
            .agg({self.valueColumn + '_end': 'sum'})
            .merge(end_df[group_by_columns], on=group_by_columns)
            .rename(columns={self.labelsColumn + '_x': self.labelsColumn})
            .drop_duplicates()
        )
        # we join the result to compare them
        merged_df = start_df.merge(end_df, on=self.get_join_key())
        merged_df[_RESULT_COLUMN] = (
            merged_df[f'{self.valueColumn}_end'] - merged_df[f'{self.valueColumn}_start']
        )
        merged_df = merged_df.drop(
            columns=[
                f'{self.valueColumn}_start',
                f'{self.valueColumn}_end',
            ]
        )

        # if there is a parent column, we need to aggregate for them
        if self.parentsColumn is not None:
            parents_results = merged_df.groupby(
                self.groupby + [self.parentsColumn], as_index=False
            ).agg({_RESULT_COLUMN: 'sum'})
            parents_results[self.labelsColumn] = parents_results[self.parentsColumn]
            return pd.concat([merged_df, parents_results])
        return merged_df

    # the waterfall has a very specific structure.
    # this methods create the top / bottom rows.
    # these contains the sum of values for the whole milestone, regardless of label.
    def compute_agg_milestone(self, df, start_df, milestone) -> pd.DataFrame:
        if len(self.groupby) > 0:
            groups = df[self.groupby].drop_duplicates()
            group_by = self.groupby
        else:
            groups = pd.DataFrame({'_VQB_GROUP': [0]})  # pseudo group
            start_df['_VQB_GROUP'] = 0
            groups['_VQB_GROUP'] = 0
            group_by = ['_VQB_GROUP']

        groups = groups.assign(**{self.labelsColumn: milestone})

        agg = start_df.groupby(group_by).agg({f'{self.valueColumn}': 'sum'})
        agg = (
            groups.merge(agg, on=group_by)
            .sort_values(
                by=self.valueColumn if self.sortBy == 'value' else self.labelsColumn,
                ascending=self.order == 'asc',
            )
            .rename(columns={self.labelsColumn: LABEL_WATERFALL_COLUMN})
        )

        if self.parentsColumn is not None:
            agg[GROUP_WATERFALL_COLUMN] = agg[LABEL_WATERFALL_COLUMN].astype(str)
        agg[TYPE_WATERFALL_COLUMN] = None
        agg[LABEL_WATERFALL_COLUMN] = agg[LABEL_WATERFALL_COLUMN].astype(str)
        if len(self.groupby) == 0:
            del start_df['_VQB_GROUP']
            del agg['_VQB_GROUP']
        return agg

    # this method shapes the merged DF so it matches the waterfall format
    def compute_mid(self, merged_df, df):
        result_df = DataFrame(
            {
                LABEL_WATERFALL_COLUMN: merged_df.sort_values(
                    by=self.get_sort_column(), ascending=self.order == 'asc'
                )[self.labelsColumn].astype(str)
            }
        )
        result_df[self.groupby] = merged_df.sort_values(
            by=self.get_sort_column(), ascending=self.order == 'asc'
        )[self.groupby]

        if self.parentsColumn is None:
            result_df[TYPE_WATERFALL_COLUMN] = 'Parent'
        else:
            result_df[TYPE_WATERFALL_COLUMN] = numpy.where(
                result_df['LABEL_waterfall'].isin(df[self.labelsColumn]), ['child'], ['parent']
            )
            result_df[GROUP_WATERFALL_COLUMN] = merged_df.sort_values(
                by=self.get_sort_column(), ascending=self.order == 'asc'
            )[self.parentsColumn]

        result_df[self.valueColumn] = merged_df.sort_values(
            by=self.get_sort_column(), ascending=self.order == 'asc'
        )[_RESULT_COLUMN]
        return result_df

    def get_sort_column(self):
        if self.sortBy == 'value':
            return _RESULT_COLUMN
        else:
            return self.labelsColumn

    def get_join_key(self):
        if self.parentsColumn is None:
            return [self.labelsColumn] + self.groupby
        else:
            return [self.labelsColumn, self.parentsColumn] + self.groupby



from typing import List, Literal, Tuple, Union

from pandas import DataFrame, merge
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.steps.combination import PipelineOrDomainName, resolve_pipeline_for_combination
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor
from weaverbird.utils import rename_duplicated_columns

JoinColumnsPair = Tuple[ColumnName, ColumnName]


class JoinStep(BaseStep):
    name = Field('join', const=True)
    right_pipeline: Union[PipelineOrDomainName]
    type: Literal['left', 'inner', 'left outer']
    on: List[JoinColumnsPair] = Field(..., min_items=1)

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever,
        execute_pipeline: PipelineExecutor,
    ) -> DataFrame:
        right_df = resolve_pipeline_for_combination(
            self.right_pipeline, domain_retriever, execute_pipeline
        )

        if self.type == 'left outer':
            how = 'outer'
        else:
            how = self.type

        result = merge(
            df,
            right_df,
            left_on=[o[0] for o in self.on],
            right_on=[o[1] for o in self.on],
            how=how,
            suffixes=('', '_JOIN'),
        )
        return rename_duplicated_columns(result)



from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName


class DuplicateStep(BaseStep):
    name = Field('duplicate', const=True)
    column: ColumnName
    new_column_name: ColumnName

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        return df.assign(**{self.new_column_name: df[self.column]})



from typing import List

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.steps.combination import PipelineOrDomainName, resolve_pipeline_for_combination
from weaverbird.types import DomainRetriever, PipelineExecutor


class AppendStep(BaseStep):
    name = Field('append', const=True)
    pipelines: List[PipelineOrDomainName]

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever,
        execute_pipeline: PipelineExecutor,
    ) -> DataFrame:
        other_dfs = [
            resolve_pipeline_for_combination(
                pipeline,
                domain_retriever,
                execute_pipeline,
            )
            for pipeline in self.pipelines
        ]
        return df.append(other_dfs, ignore_index=True)



from .addmissingdates import AddMissingDatesStep
from .aggregate import AggregateStep
from .append import AppendStep
from .argmax import ArgmaxStep
from .argmin import ArgminStep
from .base import BaseStep
from .concatenate import ConcatenateStep
from .convert import ConvertStep
from .cumsum import CumSumStep
from .date_extract import DateExtractStep
from .delete import DeleteStep
from .domain import DomainStep
from .duplicate import DuplicateStep
from .duration import DurationStep
from .evolution import EvolutionStep
from .fillna import FillnaStep
from .filter import FilterStep
from .formula import FormulaStep
from .fromdate import FromdateStep
from .ifthenelse import IfthenelseStep
from .join import JoinStep
from .lowercase import LowercaseStep
from .moving_average import MovingAverageStep
from .percentage import PercentageStep
from .pivot import PivotStep
from .rank import RankStep
from .rename import RenameStep
from .replace import ReplaceStep
from .rollup import RollupStep
from .select import SelectStep
from .sort import SortStep
from .split import SplitStep
from .statistics import StatisticsStep
from .substring import SubstringStep
from .text import TextStep
from .todate import ToDateStep
from .top import TopStep
from .totals import TotalsStep
from .uniquegroups import UniqueGroupsStep
from .unpivot import UnpivotStep
from .uppercase import UppercaseStep
from .waterfall import WaterfallStep



from typing import List, Literal

from pandas import DataFrame, Series, to_datetime, to_numeric
from pydantic import Field

from weaverbird.steps.base import BaseStep

ColumnName = str


def cast_to_float(s: Series) -> Series:
    return to_numeric(s, errors='coerce')  # cast errors will result in NaN values


def cast_to_int(s: Series) -> Series:
    s = cast_to_float(s)
    # Since .astype(int) would fail on NaN values, treat them separately:
    s_integer = s[s.notna()].astype(int)
    s_nan = s[s.isna()]
    return s_integer.append(s_nan).sort_index()


def cast_to_str(s: Series) -> Series:
    return s.astype(str)


def cast_to_datetime(s: Series) -> Series:
    return to_datetime(s, errors='coerce')  # cast errors will result in NaT values


def cast_to_bool(s: Series) -> Series:
    return s.astype(bool)


CAST_FUNCTIONS = {
    'integer': cast_to_int,
    'float': cast_to_float,
    'text': cast_to_str,
    'date': cast_to_datetime,
    'boolean': cast_to_bool,
}


class ConvertStep(BaseStep):
    name = Field('convert', const=True)
    columns: List[ColumnName]
    data_type: Literal['integer', 'float', 'text', 'date', 'boolean']

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        cast_function = CAST_FUNCTIONS[self.data_type]
        transformed_columns = {col_name: cast_function(df[col_name]) for col_name in self.columns}
        return df.assign(**transformed_columns)



from typing import List

import pandas as pd
from pydantic import BaseModel, Field

from weaverbird.steps import AggregateStep, BaseStep
from weaverbird.steps.aggregate import Aggregation
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor, PopulatedWithFieldnames


class TotalDimension(BaseModel):
    class Config(PopulatedWithFieldnames):
        ...

    total_column: ColumnName = Field(alias='totalColumn')
    total_rows_label: str = Field(alias='totalRowsLabel')


class TotalsStep(BaseStep):
    name = Field('totals', const=True)
    total_dimensions: List[TotalDimension] = Field(alias='totalDimensions')
    aggregations: List[Aggregation] = Field(min_items=1)
    groups: List[ColumnName] = Field(min_items=0, default=[])

    def execute(
        self,
        df: pd.DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> pd.DataFrame:

        total_rows = []
        # for total_dimension in self.total_dimensions:
        total_rows.append(self.get_total_for_dimensions(df))

        # rename columns in the base df, so it will match with the total in schema
        col_to_rm = set()
        for aggregation in self.aggregations:
            for col, new_col in zip(aggregation.columns, aggregation.new_columns):
                if col != new_col:
                    df[new_col] = df[col]
                    col_to_rm.add(col)
        for col in col_to_rm:
            del df[col]

        result = pd.concat([df] + total_rows)
        return result

    def get_total_for_dimensions(self, df):
        group_by_columns = self.groups + [dim.total_column for dim in self.total_dimensions]

        aggregation = AggregateStep(
            name='aggregate',
            keepOriginalGranularity=False,
            aggregations=self.aggregations,
            on=group_by_columns,
        )
        aggregated_df = aggregation.execute(df)

        result_df = pd.DataFrame()
        for dimension in self.total_dimensions:
            result_df = pd.concat(
                [result_df, self.get_total_for_dimension(aggregated_df, dimension, [])]
            )
        return result_df

    def get_total_for_dimension(
        self, df, total_dimension: TotalDimension, dimensions_to_skip
    ) -> pd.DataFrame:
        # get all group_by columns: all total_dimensions, except the current one + groups
        # all columns that are either not aggregated, or groups, or total will be null
        group_by_columns = self.groups + [
            group_column.total_column
            for group_column in self.total_dimensions
            if group_column != total_dimension
        ]
        aggregations = []
        for aggregation in self.aggregations:
            agg = aggregation.copy()
            agg.columns = agg.new_columns
            aggregations.append(agg)

        aggregation = AggregateStep(
            name='aggregate',
            keepOriginalGranularity=False,
            aggregations=aggregations,
            on=group_by_columns,
        )
        aggregated_df = aggregation.execute(df)

        aggregated_df[total_dimension.total_column] = total_dimension.total_rows_label
        full_aggregation = aggregated_df.copy()
        # if we are the last dimension, we already computed all combinations
        if total_dimension != self.total_dimensions[-1]:
            for dimension in self.total_dimensions[self.total_dimensions.index(total_dimension) :]:
                if dimension not in dimensions_to_skip and dimension != total_dimension:
                    full_aggregation = pd.concat(
                        [
                            full_aggregation,
                            self.get_total_for_dimension(
                                aggregated_df, dimension, dimensions_to_skip + [dimension]
                            ),
                        ]
                    )
        return full_aggregation



from typing import List, Tuple

from pandas import DataFrame
from pydantic import Field, root_validator

from weaverbird.steps.base import BaseStep


class RenameStep(BaseStep):
    name = Field('rename', const=True)
    to_rename: List[Tuple[str, str]] = Field(..., alias='toRename')

    @root_validator(pre=True)
    def handle_legacy_syntax(cls, values):
        if 'oldname' in values and 'newname' in values:
            values['to_rename'] = [(values.pop('oldname'), values.pop('newname'))]
        return values

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        return df.rename(columns=dict(self.to_rename))



from typing import List, Literal

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class TopStep(BaseStep):
    name = Field('top', const=True)
    groups: List[ColumnName] = []
    rank_on: ColumnName
    sort: Literal['asc', 'desc']
    limit: int

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        pandas_method = 'nlargest' if self.sort == 'desc' else 'nsmallest'
        if self.groups:
            return (
                df.groupby(self.groups)
                .apply(lambda df: getattr(df, pandas_method)(self.limit, self.rank_on))
                .reset_index(drop=True)
            )
        else:
            return getattr(df, pandas_method)(self.limit, self.rank_on)



from typing import Any, List

from pandas import DataFrame
from pydantic import Field, root_validator

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class FillnaStep(BaseStep):
    name = Field('fillna', const=True)
    columns: List[ColumnName] = Field(min_items=1)
    value: Any

    @root_validator(pre=True)
    def handle_legacy_syntax(cls, values):
        if 'column' in values:
            values['columns'] = [values.pop('column')]
        return values

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        return df.fillna({col_name: self.value for col_name in self.columns})



from typing import List

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class UnpivotStep(BaseStep):
    name = Field('unpivot', const=True)
    keep: List[ColumnName]
    unpivot: List[ColumnName]
    unpivot_column_name: ColumnName
    value_column_name: ColumnName
    dropna: bool

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        df_melted = df.melt(
            id_vars=self.keep,
            value_vars=self.unpivot,
            var_name=self.unpivot_column_name,
            value_name=self.value_column_name,
        )
        return df_melted.dropna(subset=[self.value_column_name]) if self.dropna else df_melted



from typing import List, Literal, Union

import numpy as np
import pandas as pd
from pandas import Series
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor

# cf. https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
_FREQUENCIES = {'day': 'D', 'week': 'W', 'month': 'M', 'year': 'Y'}


def at_begin_period(timestamps: Series, dates_granularity: str):
    return timestamps.dt.to_period(_FREQUENCIES[dates_granularity]).dt.start_time


class AddMissingDatesStep(BaseStep):
    name = Field('addmissingdates', const=True)
    dates_column: ColumnName = Field(alias='datesColumn')
    dates_granularity: Union[
        Literal['day'], Literal['week'], Literal['month'], Literal['year']
    ] = Field(alias='datesGranularity')
    groups: List[ColumnName] = []

    def execute(
        self,
        df: pd.DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> pd.DataFrame:
        if len(self.groups) > 0:
            groups = df.groupby(self.groups, as_index=False)
        else:
            groups = [('', df)]

        result = pd.DataFrame()
        for (key, group) in groups:
            # this is used to keep the real date, if it exists, instead of the computed one by pd.Grouper
            group = group.assign(_old_date=group[self.dates_column])

            group_with_missing_dates = group.groupby(
                pd.Grouper(key=self.dates_column, freq=_FREQUENCIES[self.dates_granularity])
            ).agg('first')

            group_with_missing_dates = group_with_missing_dates.reset_index()
            group_with_missing_dates[self.groups] = key

            group_with_missing_dates[self.dates_column] = np.where(
                pd.isna(group_with_missing_dates['_old_date']),
                at_begin_period(
                    group_with_missing_dates[self.dates_column], self.dates_granularity
                ),
                group_with_missing_dates['_old_date'],
            )
            del group_with_missing_dates['_old_date']
            result = pd.concat([result, group_with_missing_dates])
        return result



from typing import Literal, Optional

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep

OPERATIONS_MAPPING = {
    'minutes': 'minute',
    'seconds': 'second',
    'dayOfYear': 'dayofyear',
}


class DateExtractStep(BaseStep):
    name = Field('dateextract', const=True)
    column: str
    operation: Literal[
        'year',
        'month',
        'day',
        'hour',
        'minutes',
        'seconds',
        'milliseconds',
        'dayOfYear',
        'dayOfWeek',
        'week',
    ]
    new_column_name: Optional[str]

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        dst_column = self.new_column_name or f'{self.column}_{self.operation}'
        serie_dt = df[self.column].dt
        if self.operation == 'week':
            result = serie_dt.isocalendar().week
        elif self.operation == 'milliseconds':
            result = serie_dt.microsecond / 1000
        elif self.operation == 'dayOfWeek':
            # result should be between 1 (sunday) and 7 (saturday)
            result = (serie_dt.dayofweek + 2) % 7
            result = result.replace({0: 7})
        else:
            operation = OPERATIONS_MAPPING.get(self.operation, self.operation)
            result = getattr(serie_dt, operation)

        return df.assign(**{dst_column: result})



from pandas import DataFrame
from pydantic import Field

from weaverbird.conditions import Condition
from weaverbird.steps.base import BaseStep


class FilterStep(BaseStep):
    name = Field('filter', const=True)
    # TODO support and/or nesting
    condition: Condition

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        return df[self.condition.filter(df)]



from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class UppercaseStep(BaseStep):
    name = Field('uppercase', const=True)
    column: ColumnName

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        return df.assign(**{self.column: df[self.column].str.upper()})



from typing import Optional

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class SubstringStep(BaseStep):
    name = Field('substring', const=True)
    column: ColumnName
    new_column_name: Optional[ColumnName] = Field(alias='newColumnName')
    start_index: int
    end_index: int

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        new_column_name = self.new_column_name or f'{self.column}_SUBSTR'
        # Weaverbird indexes start at one
        start_index = self.start_index - 1
        # Weaverbird substring end_index is inclusive. in python, it is exclusive. therefore, no need to substract one
        end_index = self.end_index

        if end_index < 0:
            end_index += 1
        if start_index < 0:
            start_index += 1

        # there is no way to get the full string with a x[start:end] when end is negative.
        if end_index == 0:
            serie = df[self.column].str[start_index:]
        else:
            serie = df[self.column].str[start_index:end_index]

        return df.assign(**{new_column_name: serie})



from typing import List

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName


class UniqueGroupsStep(BaseStep):
    name = Field('uniquegroups', const=True)
    on: List[ColumnName]

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        return df[self.on].drop_duplicates()



from typing import List

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName


class ConcatenateStep(BaseStep):
    name = Field('concatenate', const=True)
    columns: List[ColumnName] = Field(..., min_items=2)
    separator: str
    new_column_name: ColumnName

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        new_col = df[self.columns[0]].astype(str)
        for col_name in self.columns[1:]:
            new_col = new_col.str.cat(df[col_name].astype(str), sep=self.separator)
        return df.assign(**{self.new_column_name: new_col})



from typing import List, Optional

from pandas import DataFrame, concat
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName

from .aggregate import AggregateStep, Aggregation


class RollupStep(BaseStep):
    name = Field('rollup', const=True)
    hierarchy: List[ColumnName]
    # The list of columnns to aggregate, with related aggregation function to use:
    aggregations: List[Aggregation]
    # Groupby columns if rollup has to be performed by groups:
    groupby: Optional[List[ColumnName]]
    # To give a custom name to the output label column:
    label_col: Optional[ColumnName] = Field(alias='labelCol')
    # To give a custom name to the output level column:
    level_col: Optional[ColumnName] = Field(alias='levelCol')
    # To give a custom name to the output parent column:
    parent_label_col: Optional[ColumnName] = Field(alias='parentLabelCol')

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        label_col = self.label_col or 'label'
        level_col = self.level_col or 'level'
        parent_label_col = self.parent_label_col or 'parent'

        full_current_hierarchy = []
        all_results = []
        previous_level = None

        for current_level in self.hierarchy:
            full_current_hierarchy.append(current_level)
            aggregate_on_cols = (self.groupby or []) + full_current_hierarchy
            results_for_this_level = AggregateStep(
                name='aggregate',
                on=aggregate_on_cols,
                aggregations=self.aggregations,
                keepOriginalGranularity=False,
            ).execute(df)

            results_for_this_level[level_col] = current_level
            results_for_this_level[label_col] = results_for_this_level[current_level]
            if previous_level is not None:
                results_for_this_level[parent_label_col] = results_for_this_level[previous_level]

            all_results.append(results_for_this_level)
            previous_level = current_level

        columns = (
            self.hierarchy[::-1]
            + (self.groupby or [])
            + [label_col, level_col, parent_label_col]
            + sum([agg.new_columns for agg in self.aggregations], start=[])
        )  # type: ignore

        return concat(all_results)[columns]



from typing import List, Literal, Optional

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName


class RankStep(BaseStep):
    name = Field('rank', const=True)
    value_col: ColumnName = Field(alias='valueCol')
    order: Literal['asc', 'desc']
    method: Literal['standard', 'dense']
    groupby: List[ColumnName] = []
    new_column_name: Optional[ColumnName] = Field(None, alias='newColumnName')

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        new_column_name = self.new_column_name or f'{self.value_col}_RANK'
        rank_method = 'min' if self.method == 'standard' else self.method
        ascending = self.order == 'asc'
        if self.groupby:
            serie = df.groupby(self.groupby)[self.value_col]
        else:
            serie = df[self.value_col]
        rank_serie = serie.rank(method=rank_method, ascending=ascending)
        return df.assign(**{new_column_name: rank_serie}).sort_values(new_column_name)



from typing import List, Literal, Optional

import numpy as np
from pandas import DataFrame
from pydantic import BaseModel, Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName, PopulatedWithFieldnames

DUMB_GROUPBY_COLUMN_NAME = '__dumb_groupby_column_name__'

Statistics = Literal['count', 'max', 'min', 'average', 'variance', 'standard deviation']


def statistic_to_pandas_method(s: str) -> str:
    if s == 'average':
        return 'mean'
    elif s == 'variance':
        return 'var'
    elif s == 'standard deviation':
        return 'std'
    return s


class Quantile(BaseModel):
    label: Optional[str]
    nth: int
    order: int


def percentile(nth, order):
    def f(serie):
        return np.percentile(serie, nth / order * 100)

    return f


class StatisticsStep(BaseStep):
    class Config(PopulatedWithFieldnames):
        ...

    name = Field('statistics', const=True)
    column: ColumnName
    groupby_columns: List[ColumnName] = Field([], alias='groupbyColumns')
    statistics: List[Statistics]
    # Array of quantiles. Examples:
    # - median is 1rst quantile of order 2
    # - last decile is 9th quantile of order 10
    quantiles: List[Quantile]

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        groupby_columns = self.groupby_columns
        if not self.groupby_columns:
            # Just create a temporary dumb column with '1' everywhere
            # so we can group by this column:
            groupby_columns = [DUMB_GROUPBY_COLUMN_NAME]
            df = df.assign(**{DUMB_GROUPBY_COLUMN_NAME: 1})

        stat_funcs = [statistic_to_pandas_method(s) for s in self.statistics]
        stat_funcs += [percentile(q.nth, q.order) for q in self.quantiles]
        result = df.groupby(groupby_columns)[self.column].agg(stat_funcs).reset_index()

        # Now, just fix the column names:
        quantile_names = [q.label or f'{q.nth}-th {q.order}-quantile' for q in self.quantiles]
        column_names = groupby_columns + self.statistics + quantile_names  # type: ignore
        result.columns = column_names

        if not self.groupby_columns:
            # Remove the temporary column:
            del result[DUMB_GROUPBY_COLUMN_NAME]

        return result



from typing import Callable, List

from pandas import DataFrame

ColumnName = str
Pipeline = List[dict]
DomainRetriever = Callable[[str], DataFrame]
PipelineExecutor = Callable[[List[dict]], DataFrame]


class PopulatedWithFieldnames:
    allow_population_by_field_name = True



from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import DomainRetriever


class DomainStep(BaseStep):
    name = Field('domain', const=True)
    domain: str

    def execute(self, _df, domain_retriever: DomainRetriever, execute_pipeline=None) -> DataFrame:
        return domain_retriever(self.domain)



from typing import List, Literal

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep


class PivotStep(BaseStep):
    name = Field('pivot', const=True)
    index: List[str]
    column_to_pivot: str
    value_column: str
    agg_function: Literal['sum', 'avg', 'count', 'min', 'max']

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        pivoted_df = df.pivot_table(
            values=self.value_column,
            index=self.index,
            columns=self.column_to_pivot,
            aggfunc='mean' if self.agg_function == 'avg' else self.agg_function,
        ).reset_index()
        pivoted_df.columns.name = None
        return pivoted_df



from typing import List, Union

from pandas import DataFrame

from weaverbird.types import DomainRetriever, PipelineExecutor

PipelineOrDomainName = Union[List[dict], str]  # can be either a domain name or a complete pipeline


def resolve_pipeline_for_combination(
    pipeline: PipelineOrDomainName,
    domain_retriever: DomainRetriever,
    pipeline_executor: PipelineExecutor,
) -> DataFrame:
    """
    Combined pipelines can be either single domains (str), or complete pipeline (list of steps)
    """
    if isinstance(pipeline, str):
        return domain_retriever(pipeline)
    else:
        return pipeline_executor(pipeline)



from typing import Optional

from pandas import DataFrame, to_datetime
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class ToDateStep(BaseStep):
    name = Field('todate', const=True)
    column: ColumnName
    format: Optional[str]

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        datetime_serie = to_datetime(df[self.column], format=self.format, errors='coerce')
        return df.assign(**{self.column: datetime_serie})



from typing import Literal

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor

_SECOND = 1
_MINUTE = _SECOND * 60
_HOUR = _MINUTE * 60
_DAY = _HOUR * 24

DURATIONS_IN_SECOND = {'seconds': _SECOND, 'minutes': _MINUTE, 'hours': _HOUR, 'days': _DAY}


class DurationStep(BaseStep):
    name: str = Field('duration', const=True)
    new_column_name: ColumnName = Field(alias='newColumnName')
    start_date_column: ColumnName = Field(alias='startDateColumn')
    end_date_column: ColumnName = Field(alias='endDateColumn')
    duration_in: Literal['seconds', 'minutes', 'hours', 'days'] = Field(alias='durationIn')

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        duration_serie_in_seconds = (df[self.end_date_column] - df[self.start_date_column]).astype(
            'timedelta64[s]'
        )
        duration_serie_in_given_unit = (
            duration_serie_in_seconds / DURATIONS_IN_SECOND[self.duration_in]
        )
        return df.assign(**{self.new_column_name: duration_serie_in_given_unit})



import json
import logging
from typing import List

from pandas import DataFrame
from pandas.io.json import build_table_schema

from weaverbird.pipeline import Pipeline, PipelineStep
from weaverbird.types import DomainRetriever
from weaverbird.utils import StopWatch, convert_size

logger = logging.getLogger(__name__)


class PipelineExecutor:
    """
    The main class of the module.

    Its domain retriever will be use to retrieve data needed from `domains` steps.
    """

    def __init__(self, domain_retriever: DomainRetriever):
        self.retrieve_domain = domain_retriever

    def execute_pipeline(self, pipeline_steps: List[dict]) -> DataFrame:
        """
        Execute a pipeline and returns the result as a pandas DataFrame
        """
        # TODO validate the pipeline, e.g. the first step should always be a domain step
        # self.validate_pipeline()
        steps = Pipeline(steps=pipeline_steps).steps
        df = None
        stopwatch = StopWatch()
        for index, step in enumerate(steps):
            try:
                with stopwatch:
                    df = step.execute(
                        df,
                        domain_retriever=self.retrieve_domain,
                        execute_pipeline=self.execute_pipeline,
                    )
                logger.info(
                    f'step {step} used {convert_size(df.memory_usage().sum())}, took {stopwatch.interval}s to execute'
                )
            except Exception as e:
                raise PipelineExecutionFailure(step, index, e) from e
        return df

    def preview_pipeline(self, pipeline_steps: List[dict], limit: int = 50, offset: int = 0) -> str:
        """
        Execute a pipeline but returns only a slice of the results, determined by `limit` and `offset` parameters, as JSON.

        Return format follows the 'table' JSON table schema used by pandas (see
        https://pandas.pydata.org/pandas-docs/stable/user_guide/io.html#orient-options), with a few addition related to
        pagination.

        Note: it's required to use pandas `to_json` methods, as it convert NaN and dates to an appropriate format.
        """
        df = self.execute_pipeline(pipeline_steps)
        return json.dumps(
            {
                'schema': build_table_schema(df, index=False),
                'offset': offset,
                'limit': limit,
                'total': df.shape[0],
                'data': json.loads(df[offset : offset + limit].to_json(orient='records')),
            }
        )


class PipelineExecutionFailure(Exception):
    """ Raised when a error happens during the execution of the pipeline """

    def __init__(self, step: PipelineStep, index: int, original_exception: Exception):
        self.step = step
        self.index = index
        self.original_exception = original_exception
        self.message = f'Step #{index + 1} ({step.name}) failed: {original_exception}'



from typing import List, Literal, Optional

from pandas import DataFrame, DateOffset
from pydantic import Field

from weaverbird.exceptions import DuplicateError
from weaverbird.steps.base import BaseStep

EVOLUTION_TYPE = Literal['vsLastYear', 'vsLastMonth', 'vsLastWeek', 'vsLastDay']
EVOLUTION_FORMAT = Literal['abs', 'pct']


OFFSETS = {
    'vsLastYear': DateOffset(years=1),
    'vsLastMonth': DateOffset(months=1),
    'vsLastWeek': DateOffset(weeks=1),
    'vsLastDay': DateOffset(days=1),
}


class EvolutionStep(BaseStep):
    name = Field('evolution', const=True)
    date_col: str = Field(alias='dateCol')
    value_col: str = Field(alias='valueCol')
    evolution_type: EVOLUTION_TYPE = Field(alias='evolutionType')
    evolution_format: EVOLUTION_FORMAT = Field(alias='evolutionFormat')
    index_columns: List[str] = Field([], alias='indexColumns')
    new_column: Optional[str] = Field(alias='newColumn')

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        new_column = self.new_column or f'{self.value_col}_EVOL_{self.evolution_format.upper()}'

        id_cols = [self.date_col] + self.index_columns
        if df.set_index(id_cols).index.duplicated().any():
            raise DuplicateError('Multiple rows for the same date. Did you forget indexColumns?')

        date_col_offseted = df[self.date_col] + OFFSETS[self.evolution_type]
        df_offseted = df.assign(**{self.date_col: date_col_offseted})
        both = df.merge(df_offseted, on=id_cols, how='left', suffixes=(None, '_prev_date'))
        value_date, value_prev_date = both[self.value_col], both[f'{self.value_col}_prev_date']

        if self.evolution_format == 'abs':
            evolution = value_date - value_prev_date
        elif self.evolution_format == 'pct':
            evolution = value_date / value_prev_date - 1

        return df.assign(**{new_column: evolution})



from abc import ABC, abstractmethod

from pandas import DataFrame
from pydantic.main import BaseModel

from weaverbird.types import DomainRetriever, PipelineExecutor, PopulatedWithFieldnames


class BaseStep(BaseModel, ABC):
    name: str

    class Config(PopulatedWithFieldnames):
        extra = 'forbid'

    @abstractmethod
    def execute(
        self, df: DataFrame, domain_retriever: DomainRetriever, pipeline_executor: PipelineExecutor
    ) -> DataFrame:
        ...



from typing import List

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName

_TMP_GROUP_COL_NAME = '__TMP_COL_NAME'


class ArgminStep(BaseStep):
    name = Field('argmin', const=True)
    column: ColumnName
    groups: List[str] = []

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        group = self.groups

        if len(self.groups) == 0:
            # if no groups, we create a temp column with a constant in it. Grouping on it should yield a single result
            df[_TMP_GROUP_COL_NAME] = 1
            group = [_TMP_GROUP_COL_NAME]

        aggregated_df = df.groupby(group, as_index=False).agg({self.column: 'min'})

        if len(self.groups) == 0:
            # we now remove the ugly temp column that we grouped on
            del df[_TMP_GROUP_COL_NAME]
            del aggregated_df[_TMP_GROUP_COL_NAME]

        return df.merge(aggregated_df, on=[self.column] + self.groups)



from typing import Any, List, Tuple

from pandas import DataFrame
from pydantic import Field

from weaverbird.steps import BaseStep
from weaverbird.types import ColumnName, DomainRetriever, PipelineExecutor


class ReplaceStep(BaseStep):
    name = Field('replace', const=True)
    search_column: ColumnName
    to_replace: List[Tuple[Any, Any]] = Field(min_items=1)

    def execute(
        self,
        df: DataFrame,
        domain_retriever: DomainRetriever = None,
        execute_pipeline: PipelineExecutor = None,
    ) -> DataFrame:
        return df.assign(
            **{
                self.search_column: df[self.search_column].replace(
                    {old_value: new_value for (old_value, new_value) in self.to_replace}
                )
            }
        )



from typing import List, Literal, Optional

from pandas import DataFrame, concat
from pydantic import Field, root_validator
from pydantic.main import BaseModel

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName, PopulatedWithFieldnames

AggregateFn = Literal['avg', 'sum', 'min', 'max', 'count', 'first', 'last']

functions_aliases = {'avg': 'mean'}


class Aggregation(BaseModel):
    class Config(PopulatedWithFieldnames):
        ...

    new_columns: List[ColumnName] = Field(alias='newcolumns')
    agg_function: AggregateFn = Field(alias='aggfunction')
    columns: List[ColumnName]

    @root_validator(pre=True)
    def handle_legacy_syntax(cls, values):
        if 'column' in values:
            values['columns'] = [values.pop('column')]
        if 'newcolumn' in values:
            values['new_columns'] = [values.pop('newcolumn')]
        return values


def get_aggregate_fn(agg_function: str) -> str:
    if agg_function in functions_aliases:
        return functions_aliases[agg_function]
    return agg_function


def make_aggregation(aggregation: Aggregation) -> dict:
    return {
        column_name: get_aggregate_fn(aggregation.agg_function)
        for column_name in aggregation.columns
    }


class AggregateStep(BaseStep):
    name = Field('aggregate', const=True)
    on: List[ColumnName] = []
    aggregations: List[Aggregation]
    keep_original_granularity: Optional[bool] = Field(
        default=False, alias='keepOriginalGranularity'
    )

    class Config(PopulatedWithFieldnames):
        ...

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None):
        group_by_columns = self.on

        # if no group is specified, we create a pseudo column with a single value
        if len(group_by_columns) == 0:
            group_by_columns = ['__VQB__GROUP_BY__']
            df = df.assign(**{group_by_columns[0]: True})

        grouped_by_df = df.groupby(group_by_columns)
        aggregated_cols = []
        for aggregation in self.aggregations:
            for col, new_col in zip(aggregation.columns, aggregation.new_columns):
                agg_serie = (
                    grouped_by_df[col]
                    .agg(get_aggregate_fn(aggregation.agg_function))
                    .rename(new_col)
                )
                aggregated_cols.append(agg_serie)

        df_result = concat(aggregated_cols, axis=1).reset_index()

        # it is faster this way, than to transform the original df
        if self.keep_original_granularity:
            df_result = df.merge(df_result, on=group_by_columns, how='left')

        # we do not want the pseudo column to ever leave this function
        if len(self.on) == 0:
            del df_result[group_by_columns[0]]
        return df_result



class WeaverbirdError(Exception):
    """Base class for weaverbird exceptions"""


class DuplicateError(WeaverbirdError):
    """Raised when an operation encountered duplicates where it was forbidden"""



from pandas import DataFrame
from pydantic import Field

from weaverbird.steps.base import BaseStep
from weaverbird.types import ColumnName


class TextStep(BaseStep):
    name = Field('text', const=True)
    text: str
    new_column: ColumnName

    def execute(self, df: DataFrame, domain_retriever=None, execute_pipeline=None) -> DataFrame:
        return df.assign(**{self.new_column: self.text})



from abc import ABC, abstractmethod
from typing import Any, List, Literal, Union

from numpy.ma import logical_and, logical_or
from pandas import DataFrame, Series
from pydantic import BaseModel, Field

from weaverbird.types import ColumnName, PopulatedWithFieldnames


class BaseCondition(BaseModel, ABC):
    @abstractmethod
    def filter(self, df: DataFrame) -> Series:
        """
        Returns a boolean Series, that will be used to filter a DataFrame.

        Example:
            filter = condition.filter(df)
            df[filter]  # The filtered DataFrame
        """


class ComparisonCondition(BaseCondition):
    column: ColumnName
    operator: Literal['eq', 'ne', 'lt', 'le', 'gt', 'ge']
    value: Any

    def filter(self, df: DataFrame) -> Series:
        return getattr(df[self.column], self.operator)(self.value)


class InclusionCondition(BaseCondition):
    column: ColumnName
    operator: Literal['in', 'nin']
    value: List[Any]

    def filter(self, df: DataFrame) -> Series:
        f = df[self.column].isin(self.value)
        if self.operator.startswith('n'):
            return ~f
        else:
            return f


class NullCondition(BaseCondition):
    column: ColumnName
    operator: Literal['isnull', 'notnull']

    def filter(self, df: DataFrame) -> Series:
        f = df[self.column].isnull()
        if self.operator.startswith('not'):
            return ~f
        else:
            return f


class MatchCondition(BaseCondition):
    column: ColumnName
    operator: Literal['matches', 'notmatches']
    value: str

    def filter(self, df: DataFrame) -> Series:
        f = df[self.column].str.contains(self.value)
        if self.operator.startswith('not'):
            return ~f
        else:
            return f


SimpleCondition = Union[ComparisonCondition, InclusionCondition, NullCondition, MatchCondition]


class BaseConditionCombo(BaseCondition, ABC):
    class Config(PopulatedWithFieldnames):
        ...

    def to_dict(self):
        return self.dict(by_alias=True)


class ConditionComboAnd(BaseConditionCombo):
    and_: List['Condition'] = Field(..., alias='and')

    def filter(self, df: DataFrame) -> Series:
        return logical_and.reduce([c.filter(df) for c in self.and_])


class ConditionComboOr(BaseConditionCombo):
    or_: List['Condition'] = Field(..., alias='or')

    def filter(self, df: DataFrame) -> Series:
        return logical_or.reduce([c.filter(df) for c in self.or_])


Condition = Union[ConditionComboAnd, ConditionComboOr, SimpleCondition]
ConditionComboOr.update_forward_refs()
ConditionComboAnd.update_forward_refs()
